/*
 * Copyright (C) 1996,1997,1998
 *	SHUDOH Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

/* native methods for handling constantpool item */

#include <string.h>	/* for strlen(), strcmp() */

#include "NET_shudo_moba_reflect_ConstantPool.h"

#include <jni.h>	/* for native methods */
#include <native.h>	/* for native methods */
#include <monitor.h>	/* for BINCLASS_{,UN}LOCK() */
#include <utf.h>	/* for unicode2utf() */

#include "moba.h"


#define clazzConstantPool(clazz) cbConstantPool(clazz)


/*
 * Local Functions
 */
static char *cpTypeString(int type);

static int searchUtf8(ClassClass *cb, const char *utfstr);
static int searchClazz(ClassClass *cb, ClassClass *targetCb);
static int searchNameAndTypeByField(ClassClass *cb, struct fieldblock *fb);
static int searchNameAndTypeByMethod(ClassClass *cb, struct methodblock *mb);
static int searchNameAndType(ClassClass *cb, jint id);
static int searchNBytes(ClassClass *cb,
	int type/* CONSTANT_* */, char *data, int size);
static int searchNBytesNotEqual(ClassClass *cb,
	int type/* CONSTANT_* */, char *data, int size);

static struct fieldblock *findFieldByID(ClassClass *cb, jint ID);
static struct methodblock *findMethodByID(ClassClass *cb, jint ID);
static struct fieldblock *findFieldInVM(jint ID);
static struct methodblock *findMethodInVM(jint ID);


/*
 * constant pool count ֤
 */
JNIEXPORT jint JNICALL Java_NET_shudo_moba_reflect_ConstantPool_count
  (JNIEnv *env, jclass thisClazz, jclass clazz) {
  ClassClass *cb = (Hjava_lang_Class *)DeRef(env, clazz);
  return (jint)cbConstantPoolCount(cb);
}


/*
 * constantpool μ (CONSTANT_*) ֤
 */
JNIEXPORT jint JNICALL Java_NET_shudo_moba_reflect_ConstantPool_type
  (JNIEnv *env, jclass thisClazz, jclass clazz, jint index) {
  Hjava_lang_Class *clazzHandle = (Hjava_lang_Class *)DeRef(env, clazz);
  cp_item_type *constant_pool = clazzConstantPool(clazzHandle);
  unsigned char *typeTable =
	constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;

  return (jint)typeTable[index] & CONSTANT_POOL_ENTRY_TYPEMASK;
}


/*
 * (String)Ȥ Utf8 ͤ롣
 */
JNIEXPORT jstring JNICALL Java_NET_shudo_moba_reflect_ConstantPool_getUtf8Value
  (JNIEnv *env, jclass thisClazz, jclass clazz, jint index) {
  Hjava_lang_Class *clazzHandle = (Hjava_lang_Class *)DeRef(env, clazz);
  cp_item_type *constant_pool = clazzConstantPool(clazzHandle);
  char *utf = constant_pool[index].cp;
  return (*env)->NewStringUTF(env, utf);
}

/*
 * (int)ͤ롣
 */
JNIEXPORT jint JNICALL Java_NET_shudo_moba_reflect_ConstantPool_getIntValue
  (JNIEnv *env, jclass thisClazz, jclass clazz, jint index) {
  Hjava_lang_Class *clazzHandle = (Hjava_lang_Class *)DeRef(env, clazz);
  cp_item_type *constant_pool = clazzConstantPool(clazzHandle);
  return (jint)(constant_pool[index].i);
}

/*
 * (float)ͤ롣
 */
JNIEXPORT jfloat JNICALL Java_NET_shudo_moba_reflect_ConstantPool_getFloatValue
  (JNIEnv *env, jclass thisClazz, jclass clazz, jint index) {
  Hjava_lang_Class *clazzHandle = (Hjava_lang_Class *)DeRef(env, clazz);
  cp_item_type *constant_pool = clazzConstantPool(clazzHandle);
  return (jfloat)constant_pool[index].f;
}

/*
 * (long)ͤ롣
 */
JNIEXPORT jlong JNICALL Java_NET_shudo_moba_reflect_ConstantPool_getLongValue
  (JNIEnv *env, jclass thisClazz, jclass clazz, jint index) {
  Hjava_lang_Class *clazzHandle = (Hjava_lang_Class *)DeRef(env, clazz);
  cp_item_type *constant_pool = clazzConstantPool(clazzHandle);
  Java8 t1;

  t1.l = GET_INT64(t1, &constant_pool[index]);
  return (jlong)t1.l;
}

/*
 * (double)ͤ롣
 */
JNIEXPORT jdouble JNICALL Java_NET_shudo_moba_reflect_ConstantPool_getDoubleValue
  (JNIEnv *env, jclass thisClazz, jclass clazz, jint index) {
  Hjava_lang_Class *clazzHandle = (Hjava_lang_Class *)DeRef(env, clazz);
  cp_item_type *constant_pool = clazzConstantPool(clazzHandle);
  Java8 t1;

  t1.d = GET_DOUBLE(t1, &constant_pool[index]);
  return (jdouble)t1.d;
}

/*
 * (Class)ͤ롣
 */
JNIEXPORT jclass JNICALL Java_NET_shudo_moba_reflect_ConstantPool_getClassValue
  (JNIEnv *env, jclass thisClazz, jclass clazz, jint index) {
  ClassClass *cb = (Hjava_lang_Class *)DeRef(env, clazz);
  cp_item_type *constant_pool = clazzConstantPool(cb);
  unsigned char *typeTable =
	constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
  int resolved = typeTable[index] & CONSTANT_POOL_ENTRY_RESOLVED;
  ClassClass *target = NULL;

  if (resolved)
    target = constant_pool[index].clazz;
  else {
    char *name = constant_pool[constant_pool[index].i].cp;
    target = FindClassFromClass(JNIEnv2EE(env), name, TRUE, cb);
    if (!target) {
      /* not reached */
      printf("ConstantPool#getClassValue(): class %s is not found.\n", target);
      fflush(stdout);
      return 0;
    }
  }

  return MkRefLocal(env, (Hjava_lang_Class *)target, JNI_REF_HANDLE_TAG);
}

/*
 * (String)ͤ롣
 */
struct Hjava_lang_String;
JNIEXPORT jstring JNICALL Java_NET_shudo_moba_reflect_ConstantPool_getStringValue
  (JNIEnv *env, jclass thisClazz, jclass clazz, jint index) {
  ClassClass *cb = (Hjava_lang_Class *)DeRef(env, clazz);
  cp_item_type *constant_pool = clazzConstantPool(cb);
  unsigned char *typeTable =
	constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
  int resolved = typeTable[index] & CONSTANT_POOL_ENTRY_RESOLVED;

  if (resolved)
    return MkRefLocal(env, constant_pool[index].str, JNI_REF_HANDLE_TAG);
  else {
    char *utf = constant_pool[constant_pool[index].i].cp;
    return (*env)->NewStringUTF(env, utf);
  }
}


/*
 * ꤷ int ͤؤǤΥǥå롣
 */
JNIEXPORT jint JNICALL Java_NET_shudo_moba_reflect_ConstantPool_intIndex
  (JNIEnv *env, jclass thisClazz, jclass clazz, jint value) {
  ClassClass *cb = (Hjava_lang_Class *)DeRef(env, clazz);
  return searchNBytes(cb, CONSTANT_Integer, (char *)&value, sizeof(jint));
}

/*
 * ꤷ float ͤؤǤΥǥå롣
 */
JNIEXPORT jint JNICALL Java_NET_shudo_moba_reflect_ConstantPool_floatIndex
  (JNIEnv *env, jclass thisClazz, jclass clazz, jfloat value) {
  ClassClass *cb = (Hjava_lang_Class *)DeRef(env, clazz);
  return searchNBytes(cb, CONSTANT_Float, (char *)&value, sizeof(float));
}

/*
 * ꤷ long ͤؤǤΥǥå롣
 */
JNIEXPORT jint JNICALL Java_NET_shudo_moba_reflect_ConstantPool_longIndex
  (JNIEnv *env, jclass thisClazz, jclass clazz, jlong value) {
  ClassClass *cb = (Hjava_lang_Class *)DeRef(env, clazz);
  return searchNBytes(cb, CONSTANT_Long, (char *)&value, sizeof(jlong));
}

/*
 * ꤷ double ͤؤǤΥǥå롣
 */
JNIEXPORT jint JNICALL Java_NET_shudo_moba_reflect_ConstantPool_doubleIndex
  (JNIEnv *env, jclass thisClazz, jclass clazz, jdouble value) {
  ClassClass *cb = (Hjava_lang_Class *)DeRef(env, clazz);
  return searchNBytes(cb, CONSTANT_Double, (char *)&value, sizeof(double));
}

/*
 * ꤷ boolean ͤؤǤΥǥå롣
 */
JNIEXPORT jint JNICALL Java_NET_shudo_moba_reflect_ConstantPool_booleanIndex
  (JNIEnv *env, jclass thisClazz, jclass clazz, jboolean value) {
  ClassClass *cb = (Hjava_lang_Class *)DeRef(env, clazz);
  static jint intFalse = 0, intTrue = 1;

  if (value) {
#if 1	/* (int)1 Ǥ CP õ */
    return searchNBytes(cb, CONSTANT_Integer,
				(char *)&intTrue, sizeof(jint));
#else	/* 0 Ǥʤ (int)  CP õ */
    return searchNBytesNotEqual(cb, CONSTANT_Integer,
				(char *)&intFalse, sizeof(jint));
#endif
  }
  else {
    return searchNBytes(cb, CONSTANT_Integer, (char *)&intFalse, sizeof(jint));
  }
}

/*
 * ꤷ UTF8 ʸؤǤΥǥå롣
 */
JNIEXPORT jint JNICALL Java_NET_shudo_moba_reflect_ConstantPool_utf8Index
  (JNIEnv *env, jclass thisClazz, jclass clazz, jstring tgtString) {
  ClassClass *cb = (Hjava_lang_Class *)DeRef(env, clazz);
  jboolean isCopy;
  const char *utfString;
  int index;

  utfString = (*env)->GetStringUTFChars(env, tgtString, &isCopy);
  index = searchUtf8(cb, utfString);
  if (isCopy)  (*env)->ReleaseStringUTFChars(env, tgtString, utfString);

  return (jint)index;
}

/*
 * ꤷ String ؤǤΥǥå롣
 */
JNIEXPORT jint JNICALL Java_NET_shudo_moba_reflect_ConstantPool_stringIndex
  (JNIEnv *env, jclass thisClazz, jclass clazz, jstring tgtString) {
  ClassClass *cb = (Hjava_lang_Class *)DeRef(env, clazz);
  cp_item_type *constantPool = cbConstantPool(cb);
  unsigned char *typeTable = constantPool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
  int itemCount = cbConstantPoolCount(cb);

  int ret;
  int i;

  jboolean isCopy;
  const char *tgtBody;
  int tgtStrLen;


  tgtBody = (*env)->GetStringUTFChars(env, tgtString, &isCopy);
  tgtStrLen = strlen(tgtBody);

  ret = 0;
  for (i = 1; i < itemCount; i++) {
    unsigned char type = typeTable[i] & CONSTANT_POOL_ENTRY_TYPEMASK;
	/* type is CONSTANT_* */

    if (type == CONSTANT_String) {
      jboolean itemIsCopy;

      if (typeTable[i] & CONSTANT_POOL_ENTRY_RESOLVED) {
	jstring itemStr;
	const char *itemBody;

	itemStr = MkRefLocal(env, constantPool[i].str, JNI_REF_HANDLE_TAG);
	itemBody = (*env)->GetStringUTFChars(env, itemStr, &itemIsCopy);
	if (!memcmp(tgtBody, itemBody, tgtStrLen))
	  ret = i;
	if (itemIsCopy)  (*env)->ReleaseStringUTFChars(env, itemStr, itemBody);
	if (ret)
	  goto returnStringIndex;
      }
      else {
	const char *itemBody = constantPool[constantPool[i].i].cp;
	int itemBodyLen = strlen(itemBody);
	if (!memcmp(tgtBody, itemBody, tgtStrLen)) {
	  ret = i;  goto returnStringIndex;
	}
      }	/* CP entry  resolve Ƥ뤫 */
    }	/* CONSTANT_String  */
  }

returnStringIndex:
  if (isCopy)  (*env)->ReleaseStringUTFChars(env, tgtString, tgtBody);
#ifdef DEBUG
  if (ret <= 0) {
    printf("ConstantPool#stringIndex(class `%s', string `%s'):"
	 " entry is not found in CP.\n", cbName(cb), tgtBody);
    fflush(stdout);
  }
#endif
  return ret;
}

/*
 * ꤷ饹ؤǤΥǥå롣
 * null ꤵ줿 0 ֤(java.lang.Object  superclass)
 */
JNIEXPORT jint JNICALL Java_NET_shudo_moba_reflect_ConstantPool_classIndex
  (JNIEnv *env, jclass thisClazz, jclass clazz, jclass target) {
  ClassClass *classCb;
  ClassClass *targetCb;

  if (!clazz || !target)  return 0;

  classCb = (Hjava_lang_Class *)DeRef(env, clazz);
  targetCb = (Hjava_lang_Class *)DeRef(env, target);

  return searchClazz(classCb, targetCb);
}


/*
 * classfile ΥեޥåȤ constant pool Ϥ롣
 */
JNIEXPORT void JNICALL Java_NET_shudo_moba_reflect_ConstantPool_writeInClassfile
  (JNIEnv *env, jclass thisClazz, jobject out, jclass clazz) {
  ClassClass *cb = (Hjava_lang_Class *)DeRef(env, clazz);
  int cpCount = cbConstantPoolCount(cb);
  cp_item_type *constantPool = cbConstantPool(cb);
  unsigned char *typeTable = constantPool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
  int i;

#ifdef DEBUG
  printf("MetaClass#writeCPInClassfile() called.\n");  fflush(stdout);
#endif

  for (i = 1; i < cpCount; i++) {
    unsigned char type = typeTable[i] & CONSTANT_POOL_ENTRY_TYPEMASK;
    int resolved = typeTable[i] & CONSTANT_POOL_ENTRY_RESOLVED;
	/* type is CONSTANT_* */
    Java8 t1;	/* 64bit ͳǼ */

    jshort len, index0, index1;
#ifdef DEBUG
    printf("CP[%d]: %s (%d)\n", i, cpTypeString(type), type);
    fflush(stdout);
#endif

    if (type == 0)  continue;	/* ľͤ 64bit ä */


    /* tag: CONSTANT_* */
    if ((writeChar(env, out, (int)type)) < 0)  goto exc_occurred;

    switch (type) {
    case CONSTANT_Utf8:
    case CONSTANT_Unicode:
      len = strlen(constantPool[i].cp);
      if ((writeInt16(env, out, &len)) < 0)  goto exc_occurred;
      if ((writeBytes(env, out, constantPool[i].cp, (int)len)) < 0)
	goto exc_occurred;
	/* _Utf8: VM Ǥ UTF8 ΤޤޤʤΤǤΤޤ޽ */
      break;
    case CONSTANT_Integer:
    case CONSTANT_Float:
      if ((writeInt32(env, out, (jint *)&constantPool[i])) < 0)
	goto exc_occurred;
      break;
    case CONSTANT_Long:
    case CONSTANT_Double:
      if ((writeInt64(env, out, (jlong *)&constantPool[i])) < 0)
	goto exc_occurred;
      break;
    case CONSTANT_Class:
      if (resolved) {
	ClassClass *clazz = constantPool[i].clazz;
	index0 = searchUtf8(cb, cbName(clazz));
      }
      else
	index0 = constantPool[i].i;
      if ((writeInt16(env, out, &index0)) < 0)  goto exc_occurred;
      break;
    case CONSTANT_String:
      if (resolved) {
	Hjava_lang_String *hStr;
	HArrayOfChar *value;
	unicode *body;
	char *buf;

#ifdef DEBUG
	printf("constantPool[%d]: 0x%x.\n", i, *((int *)&constantPool[i]));
	fflush(stdout);
#endif
	hStr = constantPool[i].str;
	value = unhand(hStr)->value;
#ifdef DEBUG
	printf("value: 0x%x.\n", (int)value);  fflush(stdout);
#endif
	body = unhand(value)->body;
#ifdef DEBUG
	printf("body: 0x%x.\n", (int)body);  fflush(stdout);
#endif

	/* UCS2 -> UTF8 Ѵ */
	buf = unicode2utf(body, obj_length(value), (char *)NULL, 0);
	index0 = searchUtf8(cb, buf);
	free(buf);
      }
      else
	index0 = constantPool[i].i;
      if ((writeInt16(env, out, &index0)) < 0)  goto exc_occurred;
      break;
    case CONSTANT_Fieldref:
      if (resolved) {
	struct fieldblock *fb;
	fb = constantPool[i].fb;
	index0 = searchClazz(cb, fb->clazz);
	index1 = searchNameAndTypeByField(cb, fb);
	if ((writeInt16(env, out, &index0)) < 0)  goto exc_occurred;
	if ((writeInt16(env, out, &index1)) < 0)  goto exc_occurred;
      }
      else
	if ((writeInt32(env, out, (jint *)&(constantPool[i].i))) < 0)
	  goto exc_occurred;
      break;
    case CONSTANT_Methodref:
    case CONSTANT_InterfaceMethodref:
      if (resolved) {
	struct methodblock *mb;
	mb = constantPool[i].mb;
	index0 = searchClazz(cb, mb->fb.clazz);
	index1 = searchNameAndTypeByMethod(cb, mb);
#ifdef DEBUG
	printf("Class index      : %d\n", (int)index0);
	printf("NameAndType index: %d\n", (int)index1);
	fflush(stdout);
#endif
	if ((writeInt16(env, out, &index0)) < 0)  goto exc_occurred;
	if ((writeInt16(env, out, &index1)) < 0)  goto exc_occurred;
      }
      else
	if ((writeInt32(env, out, (jint *)&(constantPool[i].i))) < 0)  goto exc_occurred;
      break;
    case CONSTANT_NameAndType:
      if (resolved) {
	struct fieldblock *field;
	struct methodblock *method;
	jint id = constantPool[i].i;

#ifdef DEBUG
	printf("ID of NameAndType: %d\n", (int)id);
	fflush(stdout);
#endif
	if (field = findFieldByID(cb, id)) {
		/* 饹⤫ field õ */
	  index0 = searchUtf8(cb, field->name);
	  index1 = searchUtf8(cb, field->signature);
	}
	else if (method = findMethodByID(cb, id)) {
		/* 饹⤫ method õ */
	  index0 = searchUtf8(cb, method->fb.name);
	  index1 = searchUtf8(cb, method->fb.signature);
	}
	else if (field = findFieldInVM(id)) {
		/* JavaVM ⤫ field õ */
	  index0 = searchUtf8(cb, field->name);
	  index1 = searchUtf8(cb, field->signature);
	}
	else if (method = findMethodInVM(id)) {
		/* JavaVM ⤫ method õ */
	  index0 = searchUtf8(cb, method->fb.name);
	  index1 = searchUtf8(cb, method->fb.signature);
	}
	else {
#ifdef DEBUG
	  printf("Any field or method has ID(%d) is not found.\n", id);
	  fflush(stdout);
#endif
	  index0 = index1 = -1;
	}

	if ((writeInt16(env, out, &index0)) < 0)  goto exc_occurred;
	if ((writeInt16(env, out, &index1)) < 0)  goto exc_occurred;
      }
      else
	if ((writeInt32(env, out, (jint *)&(constantPool[i].i))) < 0)
	  goto exc_occurred;
      break;
    default:
      /* not reached. */
      /* invalid tag */
      break;
    }
  }

  return;

exc_occurred:
  {
    jclass exc_clazz = (*env)->FindClass(env, "java/io/IOException");
    (*env)->ThrowNew(env, exc_clazz,
	"ConstantPool#writeInClassfile: write failed.");
  }
}


/*
 * classfile  constant pool Хȿ롣
 */
JNIEXPORT jlong JNICALL Java_NET_shudo_moba_reflect_ConstantPool_lengthInClassfile
  (JNIEnv *env, jclass thisClazz, jclass clazz) {
  ClassClass *cb = (Hjava_lang_Class *)DeRef(env, clazz);
  int cpCount = cbConstantPoolCount(cb);
  cp_item_type *constantPool = cbConstantPool(cb);
  unsigned char *typeTable = constantPool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;

  jlong result = 0;
  int i;

  for (i = 1; i < cpCount; i++) {
    unsigned char type = typeTable[i] & CONSTANT_POOL_ENTRY_TYPEMASK;

    switch (type) {
    case CONSTANT_Utf8:
    case CONSTANT_Unicode:
      result += 3;
      result += strlen(constantPool[i].cp);
      break;
    case CONSTANT_Integer:
    case CONSTANT_Float:
      result += 5;
      break;
    case CONSTANT_Long:
    case CONSTANT_Double:
      result += 9;
      break;
    case CONSTANT_Class:
    case CONSTANT_String:
      result += 3;
      break;
    case CONSTANT_Fieldref:
    case CONSTANT_Methodref:
    case CONSTANT_InterfaceMethodref:
    case CONSTANT_NameAndType:
      result += 5;
      break;
    default:
      break;
    }
  }

  return result;
}


/*
 * Local Functions
 */

/*
 * CONSTANT_* бʸ֤
 */
static char *cpTypeString(int type) {
  static char *typeStr[] = {
	"(null)", "Utf8", "Unicode", "Integer", "Float",
	"Long", "Double", "Class", "String",
	"Fieldref", "Methodref", "InterfaceMethodref",
	"NameAndType"
  };

  if ((type < CONSTANT_Utf8) || (type > CONSTANT_NameAndType)) {
    return "(null)";
  }

  return typeStr[type];
}


/*
 * CONSTANT_Utf8 Υȥõǥå֤
 * Ĥʤ -1 ֤
 */
static int searchUtf8(ClassClass *cb, const char *utfStr) {
  int itemCount = cbConstantPoolCount(cb);
  cp_item_type *constantPool = cbConstantPool(cb);
  unsigned char *typeTable = constantPool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
  int i;

  for (i = 1; i < itemCount; i++) {
    if ((typeTable[i] & CONSTANT_POOL_ENTRY_TYPEMASK) == CONSTANT_Utf8) {
      if (!strcmp(constantPool[i].cp, utfStr))
	return i;
    }
  }

#ifdef DEBUG
  printf("searchUtf8(class `%s', utfStr `%s'):"
	 " entry is not found in CP.\n", cbName(cb), utfStr);
  fflush(stdout);
#endif
  return -1;
}

/*
 * CONSTANT_Class Υȥõǥå֤
 * Ĥʤ -1 ֤
 */
static int searchClazz(ClassClass *cb, ClassClass *targetCb) {
  int itemCount = cbConstantPoolCount(cb);
  cp_item_type *constantPool = cbConstantPool(cb);
  unsigned char *typeTable = constantPool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
  int utfIndex;
  int i;

  if ((utfIndex = searchUtf8(cb, cbName(targetCb))) < 0)
    goto searchClazzErrorReturn;

  for (i = 1; i < itemCount; i++) {
    if ((typeTable[i] & CONSTANT_POOL_ENTRY_TYPEMASK) == CONSTANT_Class) {
      if (typeTable[i] & CONSTANT_POOL_ENTRY_RESOLVED) {
	ClassClass *clazz;
	clazz = constantPool[i].clazz;
#if 0	/* 饹̾ */
	if (!strcmp(cbName(clazz), cbName(targetCb)))
#else	/* clazz Υɥ쥹 */
	if (clazz == targetCb)
#endif
	  return i;
      }
      else {
	if (!strcmp(constantPool[constantPool[i].i].cp, cbName(targetCb)))
	  return i;
      }
    }
  }

searchClazzErrorReturn:
#ifdef DEBUG
  printf("searchClazz(): class `%s' is not found in CP.\n", cbName(targetCb));
  fflush(stdout);
#endif
  return -1;
}


/*
 * (struct fieldblock *) CONSTANT_NameAndType Υȥõ
 * ǥå֤
 * Ĥʤ -1 ֤
 */
static int searchNameAndTypeByField(ClassClass *cb, struct fieldblock *fb) {
  jint id = NameAndTypeToHash(fb->name, fb->signature);
  return searchNameAndType(cb, id);
}


/*
 * (struct methodblock *) CONSTANT_NameAndType Υȥõ
 * ǥå֤
 * Ĥʤ -1 ֤
 */
static int searchNameAndTypeByMethod(ClassClass *cb, struct methodblock *mb) {
  jint id = (jint)NameAndTypeToHash(mb->fb.name, mb->fb.signature);
			/* returns (unsigned int) */
  return searchNameAndType(cb, id);
}


/*
 * NameAndTypeToHash()ǵޤ ID  CONSTANT_NameAndType Υȥõ
 * ǥå֤
 * Ĥʤ -1 ֤
 */
static int searchNameAndType(ClassClass *cb, jint id) {
  int itemCount = cbConstantPoolCount(cb);
  cp_item_type *constantPool = cbConstantPool(cb);
  unsigned char *typeTable = constantPool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
  int i;

  for (i = 1; i < itemCount; i++)
    if ((typeTable[i] & CONSTANT_POOL_ENTRY_TYPEMASK) == CONSTANT_NameAndType)
      if (constantPool[i].i == id)  return i;

#ifdef DEBUG
  printf("searchNameAndType(class `%s', id %d):"
	 " entry is not found in CP.\n", cbName(cb), id);
  fflush(stdout);
#endif
  return -1;
}


/*
 * ȥõǥå֤
 * ߤĤʤ -1 ֤
 */
static int searchNBytes(ClassClass *cb,
	int type/* CONSTANT_* */, char *data, int size) {
  int itemCount = cbConstantPoolCount(cb);
  cp_item_type *constantPool = cbConstantPool(cb);
  unsigned char *typeTable = constantPool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
  int i;

  for (i = 1; i < itemCount; i++)
    if ((typeTable[i] & CONSTANT_POOL_ENTRY_TYPEMASK) == type)
      if (!memcmp(constantPool + i, data, size))  return i;

#ifdef DEBUG
  printf("searchNBytes(class `%s', type %d, data 0x%x, size %d):"
	 " entry is not found in CP.\n", cbName(cb), type, (int)data, size);
  fflush(stdout);
#endif
  return -1;
}

/*
 * ͤפʤȥõǥå֤
 * ߤĤʤ -1 ֤
 */
static int searchNBytesNotEqual(ClassClass *cb,
	int type/* CONSTANT_* */, char *data, int size) {
  int itemCount = cbConstantPoolCount(cb);
  cp_item_type *constantPool = cbConstantPool(cb);
  unsigned char *typeTable = constantPool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
  int i;

  for (i = 1; i < itemCount; i++)
    if ((typeTable[i] & CONSTANT_POOL_ENTRY_TYPEMASK) == type)
      if (memcmp(constantPool + i, data, size))  return i;

#ifdef DEBUG
  printf("searchNBytesNotEqual(class `%s', type %d, data 0x%x, size %d):"
	 " entry is not found in CP.\n", cbName(cb), type, (int)data, size);
  fflush(stdout);
#endif
  return -1;
}


/*
 * NameAndTypeToHash()  ID  field õ
 * Ĥʤ NULL ֤
 */
static struct fieldblock *findFieldByID(ClassClass *cb, jint id) {
  struct fieldblock *field = cbFields(cb);
  int i;

  for (i = cbFieldsCount(cb); i-- > 0; field++) {
#if 1
    if (field->ID == id)
#else	/* ID 򤳤ξǷ׻ */
    if (NameAndTypeToHash(field->name, field->signature) == id)
#endif
      return field;
  }
  return NULL;
}

/*
 * NameAndTypeToHash()  ID  method õ
 * Ĥʤ NULL ֤
 */
static struct methodblock *findMethodByID(ClassClass *cb, jint id) {
  struct methodblock *method = cbMethods(cb);
  int i;

#if 0
  printf("findMethodByID(class `%s'):\n", cbName(cb));
#endif
  for (i = cbMethodsCount(cb); i-- > 0; method++) {
    struct fieldblock field = method->fb;
#if 0
    printf(" name: %s\n", field.name);
    printf(" sig : %s\n", field.signature);
#endif
#if 1
    if (field.ID == id)
#else	/* ID 򤳤ξǷ׻ */
    if (NameAndTypeToHash(field.name, field.signature) == id)
#endif
      return method;
  }
  return NULL;
}

/*
 * JavaVM ¸ߤ륯饹٤Ƥ field õ
 */
static struct fieldblock *findFieldInVM(jint ID) {
  struct fieldblock *field;
  ClassClass *clazz;
  int i;

  ClassClass **clazzArray = get_binclasses();
  long clazzCount = get_nbinclasses();

  for (i = clazzCount; i-- > 0; ) {	/* ٤ƤΥ饹é */
    clazz = clazzArray[i];

    field = findFieldByID(clazz, ID);
    if (field)  goto returnFindFieldInVM;
  }
  field = NULL;	/* not found */

returnFindFieldInVM:
  return field;
}

/*
 * JavaVM ¸ߤ륯饹٤Ƥ method õ
 */
static struct methodblock *findMethodInVM(jint ID) {
  struct methodblock *method;
  ClassClass *clazz;
  int i;

  ClassClass **clazzArray = get_binclasses();
  long clazzCount = get_nbinclasses();

  BINCLASS_LOCK();

  for (i = clazzCount; i-- > 0; ) {	/* ٤ƤΥ饹é */
    clazz = clazzArray[i];

    method = findMethodByID(clazz, ID);
    if (method)  goto returnFindMethodInVM;
  }
  method = NULL;	/* not found */

returnFindMethodInVM:
  BINCLASS_UNLOCK();
  return method;
}
