/*
 * Copyright (C) 1996,1997,1998
 *	SHUDOH Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

#include <stdio.h>	/* for NULL */
#include <stdlib.h>	/* for malloc(), calloc() */
#include <string.h>	/* for memcpy() */
#include <netinet/in.h>	/* for hton[ls],ntoh[ls]() */

#include <jni.h>	/* for native methods */

#include "moba.h"


#define MIN(A,B) (((A)<(B))?(A):(B))

/*
 * Local Functions
 */
static jbyteArray getBuffer(JNIEnv *env, jsize size);
static void flushOutputStream(JNIEnv *env,
			jobject /*java.io.OutputStream*/ out);

static void show64(char *x);


/*
 * Global Variables
 */
static jbyteArray *bufTable = NULL;
static jobject lock;
static int lockExists = 0;


/*
 * Хåե (byte[]) Ͽ롣
 */
#define LOCK_CLASSNAME "NET/shudo/moba/io/IOLock"
#define BUF_TBLSIZE 100
static jbyteArray getBuffer(JNIEnv *env, jsize size) {
  jbyteArray array;

  if (!lockExists) {	/* lockExists ˤ븷̩¾ϴԤƤʤ */
    lock = (jobject)(*env)->FindClass(env, LOCK_CLASSNAME);
    lock = (*env)->NewGlobalRef(env, lock);
    lockExists = 1;
  }

  (*env)->MonitorEnter(env, lock);
  {
    unsigned int hash;
    if (!bufTable)
      bufTable = (jbyteArray *)calloc(sizeof(jbyteArray),BUF_TBLSIZE);
    hash = (((unsigned int)env) >> 3) % BUF_TBLSIZE;
	/* (JNIEnv *)  thread Ȥ˰ۤʤ뤳ȤԤ hash ͤ */
    if (bufTable[hash])
      array = bufTable[hash];
    else {	/* buffer  */
      array = (*env)->NewByteArray(env, size);
#ifdef DEBUG
      if (!array) {
	printf("getBuffer(): Cannot allocate an array of byte.\n");
	fflush(stdout);
      }
#endif
      array = (*env)->NewGlobalRef(env, (jobject)array);
	/*  Global ref. Ƥʤ! */
      if (array)
	bufTable[hash] = array;
    }
  }
  (*env)->MonitorExit(env, lock);

  return array;
}


/*
 * Х (OutputStream) out ˽񤭽Ф
 *
 * returns: : 0, : -1
 */
#define BUF_SIZE 4000
extern int writeBytes(JNIEnv *env,
		jobject /*java.io.OutputStream*/ out, char *data, size_t len) {
  jbyteArray array;

#ifdef WRITE_VERBOSE
  printf("writeBytes() (len:%d) called.\n", len);
  fflush(stdout);
#endif

  array = getBuffer(env, BUF_SIZE);

  (*env)->MonitorEnter(env, (jobject)array);
  {
    jsize writeSize;
    while (len > 0) {
      writeSize = MIN(len, BUF_SIZE);
      (*env)->SetByteArrayRegion(env, array, 0, writeSize, (jbyte *)data);
	/* memory copy... */
      if ((writeArrayOfByte(env, out, array, 0, len)) < 0)  return -1;
	/* returns : 0, : -1 */

      data += writeSize;
      len -= writeSize;
    }
  }
  (*env)->MonitorExit(env, (jobject)array);

  return 0;
}


/*
 * Х (InputStream) in ɤ߹ࡣ
 * ǡΰѰդƤɬפϤʤ
 */
extern size_t readBytes(JNIEnv *env,
		jobject /*java.io.InputStream*/ in, char *data, size_t len) {
  jbyteArray array;
  size_t readSum;

  array = getBuffer(env, BUF_SIZE);

  (*env)->MonitorEnter(env, (jobject)array);
  {
    jsize readSize;
    jboolean isCopy;
    jbyte *readData;

    readSum = 0;
    while (len > 0) {
      readSize = MIN(len, BUF_SIZE);
      if ((readArrayOfByte(env, in, array, 0, readSize)) < 0)  return -1;
      readData = (*env)->GetByteArrayElements(env, array, &isCopy);
      memcpy(data, readData, (size_t)readSize);
      if (isCopy)
	(*env)->ReleaseByteArrayElements(env, array, readData, JNI_ABORT);

      data += readSize;
      len -= readSize;
      readSum += readSize;
    }
  }
  (*env)->MonitorExit(env, (jobject)array);

  return readSum;
}


/*
 * ObjectWriter#writeObject(Object) ƤӽФ
 */
extern int writeObject(JNIEnv *env,
		jobject /*NET.shudo.moba.io.ObjectWriter*/ out, jobject obj) {
  static int methodExists = 0;
  static jmethodID method;

  if (!methodExists) {
    jclass clazz = (*env)->FindClass(env, WRITEOBJ_CLASSNAME);
    (*env)->NewGlobalRef(env, clazz);
    method = (*env)->GetMethodID(env, clazz,
			WRITEOBJ_METHOD, WRITEOBJ_SIGNATURE);
    if ((*env)->ExceptionOccurred(env)) {
      /* not reached. */
      printf("method " WRITEOBJ_SIGNATURE " " WRITEOBJ_METHOD "
		is not found.\n");
      fflush(stdout);
      return;
    }
    else  methodExists = 1;
  }

  (*env)->CallVoidMethod(env, out, method, obj);

  if ((*env)->ExceptionOccurred(env)) {
#ifdef DEBUG
    printf("writeObject(): An exception occurred.\n");
    fflush(stdout);
#endif
    return -1;
  }

  return 0;
}


/*
 * ObjectReader#readObject() ƤӽФ
 */
extern jobject readObject(JNIEnv *env,
		jobject /*NET.shudo.moba.io.ObjectReader*/ in) {
  static int methodExists = 0;
  static jmethodID method;

  jobject result = NULL;

  if (!methodExists) {
    jclass clazz = (*env)->FindClass(env, READOBJ_CLASSNAME);
    (*env)->NewGlobalRef(env, clazz);
    method = (*env)->GetMethodID(env, clazz,
			READOBJ_METHOD, READOBJ_SIGNATURE);
    if ((*env)->ExceptionOccurred(env)) {
      /* not reached. */
      printf("method " READOBJ_SIGNATURE " " READOBJ_METHOD "
		is not found.\n");
      fflush(stdout);
      return;
    }
    else  methodExists = 1;
  }

  result = (*env)->CallObjectMethod(env, in, method);

  if ((*env)->ExceptionOccurred(env)) {
#ifdef DEBUG
    printf("readObject(): An exception occurred.\n");
    (*env)->ExceptionDescribe(env);
    fflush(stdout);
#endif
    return result;
  }

  return result;
}


/*
 * 64bit 񤭹ࡣ
 */
extern int writeInt64(JNIEnv *env,
		jobject /*java.io.OutputStream*/ out, jlong *p) {
#ifdef BIG_ENDIAN
  return writeBytes(env, out, (char *)p, 8);
#else
  jlong v = hton64(*p);
  return writeBytes(env, out, (char *)&v, 8);
#endif
}


/*
 * 32bit 񤭹ࡣ
 */
extern int writeInt32(JNIEnv *env,
		jobject /*java.io.OutputStream*/ out, jint *p) {
#ifdef BIG_ENDIAN
  return writeBytes(env, out, (char *)p, 4);
#else
  jint v;
  v = (jint)htonl(*(unsigned long *)p);
  return writeBytes(env, out, (char *)&v, 4);
#endif
}


/*
 * 16bit 񤭹ࡣ
 */
extern int writeInt16(JNIEnv *env,
		jobject /*java_io_OutputStream*/ out, jshort *p) {
#ifdef BIG_ENDIAN
  return writeBytes(env, out, (char *)p, 2);
#else
  jshort v;
  v = (jshort)htons(*(unsigned short *)p);
  return writeBytes(env, out, (char *)&v, 2);
#endif
}


/*
 * char 񤭹ࡣ
 */
extern int writeChar(JNIEnv *env,
		jobject /*java.io.OutputStream*/ out, int c) {
  char ch = (char)c;
  return writeBytes(env, out, &ch, 1);
}


static void show64(char *x) {
  int i;

  printf("0x%02x", *x++);
  for (i = 1; i < 8; i++) {
    printf(", 0x%02x", *x++);
  }
  printf("\n");
  fflush(stdout);
}
/*
 * 64bit ɤ߹ࡣ
 */
extern jlong readInt64(JNIEnv *env, jobject /*java.io.InputStream*/ in) {
  jlong v;
  if (readBytes(env, in, (char *)&v, 8) != 8)
    return -1;
#ifdef BIG_ENDIAN
  return v;
#else
  return ntoh64(v);
#endif
}


/*
 * 32bit ɤ߹ࡣ
 */
extern jint readInt32(JNIEnv *env, jobject /*java.io.InputStream*/ in) {
  jint ret = 0;

  if (readBytes(env, in, (char *)&ret, 4) != 4)
    return -1;
#ifdef BIG_ENDIAN
  return ret;
#else
  return (jint)ntohl(*(unsigned long *)&ret);
#endif
}


/*
 * 16bit ɤ߹ࡣ
 */
extern jshort readInt16(JNIEnv *env, jobject /*java_io_InputStream*/ in) {
  jshort ret = 0;

  if (readBytes(env, in, (char *)&ret, 2) != 2)
    return -1;
#ifdef BIG_ENDIAN
  return ret;
#else
  return (jshort)ntohs(*(unsigned short *)&ret);
#endif
}


/*
 * char ɤ߹ࡣ
 */
extern jbyte readChar(JNIEnv *env, jobject /*java.io.InputStream*/ in) {
  char ret;

  if (readBytes(env, in, &ret, 1) != 1)
    return -1;
  return ret;
}


/*
 * ʸĹ, ʸ񤭽Ф
 *
 * returns: : 0, : -1
 */
extern int writeString(JNIEnv *env,
		jobject /*java.io.OutputStream*/ out, char *str) {
  size_t len = strlen(str);
  jshort len16 = (jshort)len;

  /* ʸĹ */
  if ((writeInt16(env, out, &len16)) < 0)  return -1;

  /* ʸ */
  if ((writeBytes(env, out, str, len)) < 0)  return -1;

  return 0;
}


/*
 * ʸĹ, ʸɤ߹ࡣ
 */
extern char *readStrdup(JNIEnv *env, jobject /*java.io.InputStream*/ in) {
  jshort len16;
  size_t len;
  char *ptr;

  /* ʸĹ */
  len16 = readInt16(env, in);
  len = len16 | (size_t)0;

  if (!(ptr = (char *)malloc(len + 1))) {
    SignalError(0, JAVAPKG "OutOfMemoryError", 0);
    return NULL;
  }

  /* ʸ */
  if (readBytes(env, in, ptr, len) != len)
    return NULL;
  ptr[len] = '\0';	/* terminate */

  return ptr;
}


#define WRITEBYTE_CLASSNAME	"java/io/OutputStream"
#define WRITEBYTE_SIGNATURE	"([BII)V"
#define WRITEBYTE_METHOD	"write"
#define READBYTE_CLASSNAME	"java/io/InputStream"
#define READBYTE_SIGNATURE	"([BII)I"
#define READBYTE_METHOD		"read"
#define FLUSH_CLASSNAME		"java/io/OutputStream"
#define FLUSH_SIGNATURE		"()V"
#define FLUSH_METHOD		"flush"


/*
 * OutputStream#write(byte[], int, int) ƤӽФ
 */
jint writeArrayOfByte(JNIEnv *env, jobject /*java.io.OutputStream*/ out,
	jbyteArray data, jint off, jint len) {
  static int methodExists = 0;
  static jmethodID method;
#ifdef WRITE_VERBOSE
  printf("writeArrayOfByte() called.\n");
  fflush(stdout);
#endif

  if (!methodExists) {
    jclass clazz = (*env)->FindClass(env, WRITEBYTE_CLASSNAME);
    (*env)->NewGlobalRef(env, clazz);
    method = (*env)->GetMethodID(env, clazz,
			WRITEBYTE_METHOD, WRITEBYTE_SIGNATURE);
    if ((*env)->ExceptionOccurred(env)) {
      /* not reached. */
      printf("method " WRITEBYTE_SIGNATURE " " WRITEBYTE_METHOD " is not found.\n");
      fflush(stdout);
      return -1;
    }
    else  methodExists = 1;
  }

  (*env)->CallVoidMethod(env, out, method, data, off, len);

  if ((*env)->ExceptionOccurred(env)) {
    /* (*env)->ExceptionClear(env); */
#ifdef DEBUG
    printf("writeArrayOfByte(): An exception occurred.\n");
    fflush(stdout);
#endif
    return -1;
  }
  else
    return 0;
}


/*
 * InputStream#read(byte[]) ƤӽФ
 *
 * returns: ɤ᤿Хȿ, Ԥ -1
 */
int readArrayOfByte(JNIEnv *env, jobject /*java.io.InputStream*/ in,
	jbyteArray data, jint off, jint len) {
  static int methodExists = 0;
  static jmethodID method;
  jint ret;
  jint readLen;

  if (!methodExists) {
    jclass clazz = (*env)->FindClass(env, READBYTE_CLASSNAME);
    (*env)->NewGlobalRef(env, clazz);
    method = (*env)->GetMethodID(env, clazz,
			READBYTE_METHOD, READBYTE_SIGNATURE);
    if ((*env)->ExceptionOccurred(env)) {
      /* not reached. */
      printf("method " READBYTE_SIGNATURE " " READBYTE_METHOD " is not found.\n");
      fflush(stdout);
      return -1;
    }
    else  methodExists = 1;
  }

  readLen = 0;
  while (1) {
    ret = (*env)->CallIntMethod(env, in, method, data, off, len);

#ifdef WRITE_VERBOSE
    {
      jboolean isCopy;
      jbyte *body = (*env)->GetByteArrayElements(env, data, &isCopy);
      jsize bodylen = (*env)->GetArrayLength(env, data);
      int i;
      printf("readArrayOfByte(): ((byte[%d])data, %d, %d) returns %d.\n",
		bodylen, off, len, ret);
      for (i = 0; i < ret; i++)  printf(" %02x", (int)body[off + i] & 0xff);
      if (isCopy)
	(*env)->ReleaseByteArrayElements(env, data, body, JNI_ABORT);
      printf("\n");  fflush(stdout);
    }
#endif

    if ((*env)->ExceptionOccurred(env)) {
#ifdef DEBUG
      printf("readArrayOfByte(): An exception occurred.\n");
      fflush(stdout);
#endif
      return -1;
    }

    if (ret < 0)  break;	/* EOF */

    len -= ret;  off += ret;
    readLen += ret;
    if (len <= 0)  break;
  }

  return readLen;
}


/*
 * Local Functions
 */

/*
 * OutputStream#flush() ƤӽФ
 */
static void flushOutputStream(JNIEnv *env, jobject /*java.io.OutputStream*/ out) {
  static int methodExists = 0;
  static jmethodID method;

  if (!methodExists) {
    jclass clazz = (*env)->FindClass(env, FLUSH_CLASSNAME);
    (*env)->NewGlobalRef(env, clazz);
    method = (*env)->GetMethodID(env, clazz, FLUSH_METHOD, FLUSH_SIGNATURE);
    if ((*env)->ExceptionOccurred(env)) {
      /* not reached. */
      printf("method " FLUSH_SIGNATURE " " FLUSH_METHOD " is not found.\n");
      fflush(stdout);
      return;
    }
    else  methodExists = 1;
  }

  (*env)->CallVoidMethod(env, out, method);
#ifdef DEBUG
  if ((*env)->ExceptionOccurred(env)) {
    printf("flushOutputStream(): An exception occurred.\n");
    fflush(stdout);
  }
#endif
}
