/*
 * Copyright (C) 1996,1997,1998
 *	SHUDOH Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

package NET.shudo.moba.place;


import java.util.Dictionary;
import java.util.Hashtable;
import java.net.Socket;
import java.net.InetAddress;
import java.io.IOException;
import java.net.UnknownHostException;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import NET.shudo.moba.lang.PlaceAddress;


/**
 * ưåɤΤΥ饹<BR>
 * java.*, sun.*, NET.shudo.moba.* ʳΥ饹
 * åɤθε꤫ͥåȥͳǥɤ롣<BR>
 * Ȥб ClassDistributor 饯饹(饹ե)롣
 *
 * @see NET.shudo.moba.place.ClassDistributor
 */
public final
class MobaClassLoader extends ClassLoader implements ReferenceCounter {
  /**
   * MobaClassLoaderɽ饹Υɸɽ InetAddress 
   */
  private static Dictionary classLoaderTable = new Hashtable();

  /**
   * debug flag
   */
  private boolean debug = false;

  /**
   * Υ饹ͭɤ
   *
   * @see unregister()
   * @see init()
   */
  private boolean enabled = false;

  /**
   * ɸʳΥ饹˥ͥåȥͳǥɤ뤫
   */
  private boolean alwaysLoadNotStandardClassesViaNetwork = true;

  /**
   * ³衢Ĥޤꥯ饹ɸΥɥ쥹
   */
  PlaceAddress peerAddress;

  /**
   * 饹ФؤΥͥ
   */
  Socket sock = null;

  DataInputStream in = null;
  DataOutputStream out = null;


  /**
   * 󥹥ȥ饯<BR>
   * 桼ץफƤ֤٤ǤϤʤget() Ѥ٤
   *
   * @see NET.shudo.moba.place.MobaClassLoader#get()
   */
  private MobaClassLoader(PlaceAddress addr) throws IOException {
    this.init(addr);
  }


  /**
   * MobaClassLoader Υ󥹥󥹤롣<BR>
   * 󥹥ȥ饯Ѥ٤Ǥʤ
   */
  public static MobaClassLoader get(PlaceAddress addr) throws IOException {
    MobaClassLoader cl = null;

    synchronized (classLoaderTable) {
      cl = (MobaClassLoader)classLoaderTable.get(addr);
      if (cl != null) {
	if (!cl.enabled)  cl.init(addr);
      }
      else {
	cl = new MobaClassLoader(addr);
	classLoaderTable.put(addr, cl);
      }
    }

    return cl;
  }
  public static MobaClassLoader get(InetAddress addr, int port)
	throws IOException {
    return get(new PlaceAddress(addr, port));
  }
  public static MobaClassLoader get(InetAddress addr) throws IOException {
    return get(new PlaceAddress(addr));
  }
  public static MobaClassLoader get(String host, int port) throws IOException {
    InetAddress addr = InetAddress.getByName(host);
    return get(new PlaceAddress(addr, port));
  }
  public static MobaClassLoader get(String host) throws IOException {
    return get(new PlaceAddress(host));
  }


  /**
   * 
   */
  private synchronized void init(PlaceAddress addr) throws IOException {
    this.peerAddress = addr;

    sock = new Socket(addr.inetAddress(),
			addr.port() + PlaceAddress.CLASS_PORT_DIFF);
    in = new DataInputStream(sock.getInputStream());
    out = new DataOutputStream(sock.getOutputStream());

    this.enabled = true;
  }


  /**
   * ȥȡ
   */
  private int refCount = 0;

  /**
   * ȥȤ䤹
   *
   * @see unregister()
   */
  public void register() { refCount++; }

  /**
   * ȥȤ򸺤餹
   * Υ饹ȤʤȤȤƸƤ֡<BR>
   * MetaClass Υơ֥뤫λȤĤꡢMobaClassLoader Υ󥹥󥹤
   * ʤᡢ饹¦ ClassDistributor 뤿
   * Υ᥽åɤƤ֡
   *
   * @see regeister()
   */
  public synchronized void unregister() {
    refCount--;

    if (refCount <= 0) {
      try {
	sock.close();
      }
      catch (IOException e) {
	// ̵
      }

      sock = null;  in = null;  out = null;

      this.enabled = false;
    }
  }


  /**
   * debug flag ꤹ롣
   */
  public void debug(boolean flag) { this.debug = flag; }


  /**
   * implements java.lang.ClassLoader#loadClass().
   */
  public synchronized Class loadClass(String name, boolean resolve)
	throws ClassNotFoundException {
    Class c = null;
    boolean done = false;

    if (debug) {
      System.out.println("MobaClassLoader#loadClass() classname: " + name);
      System.out.flush();
    }

    // ̵Ƥ
    if (!this.enabled) {
      try { this.init(this.peerAddress); }
      catch (IOException e) {
	throw new ClassNotFoundException(
		"initialization of class loader failed.");
      }
    }

    if ((name == null) || (name.length() == 0))
      throw new ClassNotFoundException("class name is not specified.");

    c = findLoadedClass(name);
    if (c == null) {
      if ( (!alwaysLoadNotStandardClassesViaNetwork) ||
	name.startsWith("sun.") || name.startsWith("java.") ||
	name.startsWith("[") || name.startsWith("NET.shudo.moba.") ) {
	// sun.*, java.*, NET.shudo.moba.*, Ǥ

	if (debug) {
	  System.out.println("try to load from local system.");
	  System.out.flush();
	}

	try {
	  // local system ˥ɤǤ
	  c = findSystemClass(name);
	  done = true;
	}
	catch (ClassNotFoundException e) {}
      }

      if (!done) {
	if (debug) {
	  System.out.println("try to load from network.");
	  System.out.flush();
	}

	if (sock == null) {
	  throw new ClassNotFoundException(
		"this classloader is already disabled.");
	}

	byte[] data = null;
	int retryCount = 0;
	do {
	  // ͥåȥͳΥɤߤ
	  try {
	    out.writeUTF(name);

	    int dataLength = in.readInt();
	    if (debug) {
	      System.out.println("length: " + dataLength);  System.out.flush();
	    }
	    if (dataLength <= 0)  throw new ClassNotFoundException(name);
	    data = new byte[dataLength];

	    int remain = data.length;
	    while (remain > 0) {
	      remain -= in.read(data, data.length - remain, remain);
	    }

	    done = true;
	  }
	  catch (IOException e) {
	    if (retryCount > 0)
	      throw new ClassNotFoundException(name);

	    // ľ
	    retryCount++;
	    try {  this.init(this.peerAddress);  }
	    catch (IOException initEx) {
	      throw new ClassNotFoundException(name);
	    }
	  }
	} while (!done);
/*
try {
  java.io.OutputStream out = new java.io.FileOutputStream("classfile." + name);
  out.write(data);
  out.close();
}
catch (IOException e) {
  e.printStackTrace();
}
*/

	c = defineClass(name, data, 0, data.length);
      }
    }	// if (c == null)

    if (resolve)
      resolveClass(c);

    if (debug) {
      System.out.println("MobaClassLoader#loadClass() done.");
      System.out.flush();
    }
    return c;
  }
}
