/*
 * Copyright (C) 1996,1997,1998
 *	SHUDOH Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

package NET.shudo.moba.place;


import java.net.Socket;
import java.net.ServerSocket;
import java.io.OutputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import NET.shudo.moba.lang.PlaceAddress;
// native methods ΥɤΤ
import NET.shudo.moba.lang.MobaThread;
// place  InetAddress Τ
import java.net.InetAddress;
import java.net.UnknownHostException;
// Τ
import NET.shudo.moba.net.AccessControler;


/**
 * åɤ place<BR>
 * ThreadServer, ClassServer, CommandServer ʤ
 * Ƽǡư롣
 *
 * @see NET.shudo.moba.place.ThreadServer
 * @see NET.shudo.moba.place.ClassServer
 * @see NET.shudo.moba.place.CommandServer
 */
public
class MobaPlace {
  private static final String PROGRAM = "mobaplace";

  /**
   * ǥХåե饰
   */
  private boolean debug = false;

  /**
   * åɤ ThreadGroup ̾
   */
  public static final String THREAD_POOL_NAME = "Thread Pool";

  /**
   * åɤ ThreadGroup
   */
  public static final ThreadGroup threadPool =
		new ThreadGroup(THREAD_POOL_NAME);

  /**
   * ȥ顣
   */
  public static AccessControler accessControler = null;


  public MobaPlace() {
  }


  public void usage() {
    System.out.println("usage: " + this.PROGRAM + " [-h] [-d] [-p port] [-a access_file]");
    System.out.println("\t-h:        Show this help message.");
    System.out.println("\t-p port:   Wait on specified port (a place require 3 ports).");
    System.out.println("\t-a file:   Enable access control with specified control file.");
    System.out.println("\t-l file:   Do logging to specified log file.");
    System.out.println("\t-n number: Exit after received specified numbers of threads.");
    System.out.println("\t-d:        For debugging.");
  }


  private void start(String[] argv) {
    int maxReceipt = 0;	// means infinity

    debug = false;

    OutputStream logOut = null;	// ν


    // parse options
    for (int i = 0; i < argv.length; i++) {
      if (argv[i].equals("-d")) {
	this.debug = true;
      }
      else if (argv[i].equals("-h") || argv[i].equals("-?")) {
	usage();  System.exit(0);
      }
      else if (argv[i].equals("-a")) {
	if ((i + 1) >= argv.length) {
	  System.err.println("Option -a requires an argument(file name).");
	  usage();  System.exit(1);
	}
	String accessFilename = argv[++i];
	try {
	  accessControler = new AccessControler(accessFilename, debug);
	}
	catch (IOException e) {
	  System.err.println("Cannot open or fail in reading access control file: " + accessFilename);
	  System.exit(1);
	}
      }
      else if (argv[i].equals("-l")) {
	if ((i + 1) >= argv.length) {
	  System.err.println("Option -l requires an argument(file name).");
	  usage();  System.exit(1);
	}
	String logFilename = argv[++i];
	try {
	  logOut = new FileOutputStream(logFilename);
	}
	catch (IOException e) {
	  System.err.println("Cannot open log file: " + logFilename);
	  System.exit(1);
	}
      }
      else if (argv[i].equals("-p")) {
	if ((i + 1) >= argv.length) {
	  System.err.println("Option -p requires an argument(port no.).");
	  usage();  System.exit(1);
	}
	PlaceAddress.PLACE_PORT = Integer.parseInt(argv[++i]);
      }
      else if (argv[i].equals("-n")) {
	if ((i + 1) >= argv.length) {
	  System.err.println("Option -n requires an argument(number).");
	  usage();  System.exit(1);
	}
	maxReceipt = Integer.parseInt(argv[++i]);
	if (debug) {
	  System.out.println("max receipts of threads: " + maxReceipt);
	}
      }
      else if (argv[i].startsWith("-", 0)) {
	System.err.println("Invalid option: " + argv[i]);
	usage();  System.exit(1);
      }
//      else if (argv[i].equals()) {
//      }
      else {	// 
      }
    }


    // 
    MobaThread.loadLibrary();


    // Ƽǡư
    Runnable r;
    Thread t;

    // Logger ư
    if (logOut != null) {
      r = new Logger(logOut);
      t = new Thread(r, "Logger");
      t.setDaemon(true);	// daemon thread
      try {
	t.setPriority(Thread.MIN_PRIORITY);
      } catch (IllegalArgumentException e) {}
      t.start();
    }

    // ThreadServer ư
    r = new ThreadServer(PlaceAddress.PLACE_PORT, debug, maxReceipt);
    t = new Thread(r, "Thread Server");
    t.setDaemon(false);	// user thread
    try {
      t.setPriority(Thread.NORM_PRIORITY);
    } catch (IllegalArgumentException e) {}
    t.start();

    // ClassServer ư
    r = new ClassServer(PlaceAddress.PLACE_PORT, debug);
    t = new Thread(r, "Class Server");
    t.setDaemon(true);	// daemon thread
    try {
      t.setPriority(Thread.NORM_PRIORITY + 1);
    } catch (IllegalArgumentException e) {}
    t.start();

    // CommandServer ư
    r = new CommandServer(PlaceAddress.PLACE_PORT, debug);
    t = new Thread(r, "Command Server");
    t.setDaemon(true);	// daemon thread
    try {
      t.setPriority(Thread.NORM_PRIORITY + 2);
    } catch (IllegalArgumentException e) {}
    t.start();


    // boot up message
    PlaceAddress here = PlaceAddress.localAddress();
    System.out.print("MOBA place started");
    if (here != null) {
      System.out.println(" on "
	+ here.hostName() + "(" + here.hostAddress() + "):"
	+ here.port() + ".");
      System.out.println("Command interpreter is waiting on port "
	+ PlaceAddress.PLACE_PORT + ".");
    }
    else
      System.out.println();
    System.out.println("Enjoy!");
  }


  public static void main(String[] argv) {
    System.runFinalizersOnExit(true);

    new MobaPlace().start(argv);
  }
}
