/*
 * Copyright (C) 1996,1997,1998
 *	SHUDOH Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

package NET.shudo.moba.reflect;

import java.io.OutputStream;
import java.util.Dictionary;
import java.util.Hashtable;
import java.util.Enumeration;


abstract public
class ConstantPool {
  public static final int CONSTANT_UNKNOWN =	0;
  public static final int CONSTANT_Utf8 =	1;
  public static final int CONSTANT_Unicode =	2;
  public static final int CONSTANT_Integer =	3;
  public static final int CONSTANT_Float =	4;
  public static final int CONSTANT_Long =	5;
  public static final int CONSTANT_Double =	6;
  public static final int CONSTANT_Class =	7;
  public static final int CONSTANT_String =	8;
  public static final int CONSTANT_Fieldref =	9;
  public static final int CONSTANT_Methodref =	10;
  public static final int CONSTANT_InterfaceMethodref =	11;
  public static final int CONSTANT_NameAndType =	12;


  public static native int count(Class clazz);

  public static native int type(Class clazz, int index);

  protected static native String getUtf8Value(Class clazz, int index);
  protected static native int getIntValue(Class clazz, int index);
  protected static native float getFloatValue(Class clazz, int index);
  protected static native long getLongValue(Class clazz, int index);
  protected static native double getDoubleValue(Class clazz, int index);
  protected static native Class getClassValue(Class clazz, int index);
  protected static native String getStringValue(Class clazz, int index);

  public static native int intIndex(Class clazz, int value);
  public static native int floatIndex(Class clazz, float value);
  public static native int longIndex(Class clazz, long value);
  public static native int doubleIndex(Class clazz, double value);
  public static native int booleanIndex(Class clazz, boolean value);
  public static native int utf8Index(Class clazz, String utf8);
  public static native int stringIndex(Class clazz, String str);
  public static native int classIndex(Class clazz, Class target);


  /**
   * classfile ΥեޥåȤ constant pool Ϥ롣
   */
  public static native void writeInClassfile(OutputStream out, Class clazz)
	throws java.io.IOException;

  /**
   * classfile  constant pool Хȿ롣
   */
  public static native long lengthInClassfile(Class clazz);


  /**
   * constant pool ˺ߤ륯饹̾󤹤롣
   */
  protected static String[] classnames(Class clazz) {
    int count = count(clazz);
    int i, j;

    Dictionary cnameDic = new Hashtable();
    String[] cnameArray = null;
    String cname = null;

    for (i = 0; i < count; i++) {
      int type = type(clazz, i);
      switch (type) {
      case CONSTANT_Utf8:
	cnameArray = classnamesInSignature(getUtf8Value(clazz, i));
	for (j = 0; j < cnameArray.length; j++) {
	  cname = cnameArray[j];
	  cnameDic.put(cname, cname);
	}
	break;
      case CONSTANT_Class:
	cname = getClassValue(clazz, i).getName();
	cnameDic.put(cname, cname);
	break;
      default:
	break;
      }
    }

    cnameArray = new String[cnameDic.size()];
    i = 0;
    for (Enumeration e = cnameDic.elements(); e.hasMoreElements(); ) {
      cnameArray[i++] = (String)e.nextElement();
    }

/*
    // for debug
    System.out.println("ConstantPool#classnames():");
    for (i = 0; i < cnameArray.length; i++) {
      System.out.println("{" + cnameArray[i] + "}");
    }
    System.out.flush();
*/

    return cnameArray;
  }

  /**
   * signature 椫饯饹̾ȴФ
   */
  protected static String[] classnamesInSignature(String sig) {
    Dictionary ret = new Hashtable();
    String classname = null;
    int offset = 0;
    int indexOfL = 0;
    int indexOfSemicolon = 0;

//System.out.println("{" + sig + "}");
    while (true) {
      indexOfL = sig.indexOf('L', offset);
      if (indexOfL < 0)  break;
      offset = indexOfL + 1;

      indexOfSemicolon = sig.indexOf(';', offset);
      if (indexOfSemicolon < 0)  continue;
      offset = indexOfSemicolon + 1;

      classname = sig.substring(indexOfL + 1, indexOfSemicolon);
//System.out.println("(" + indexOfL + "," + indexOfSemicolon + ") {" + classname + "}");

      classname = classname.replace('/', '.');
      ret.put(classname, classname);
    }

    int retSize = ret.size();
    if (retSize > 0) {
      String[] array = new String[retSize];
      int i = 0;
      for (Enumeration e = ret.keys(); e.hasMoreElements(); ) {
	array[i++] = (String)e.nextElement();
      }
      return array;
    }
    else
      return new String[0];
  }
}
