/*
 * Copyright (C) 1996,1997,1998
 *	SHUDOH Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

package NET.shudo.moba.reflect;

import java.lang.reflect.Field;	// for init()
import java.lang.reflect.Method;	// for init(), methodCountWithAccess()
import java.lang.reflect.Constructor;	// for init()
import java.io.OutputStream;	// for write*InClassfile()
import java.io.IOException;	// for write,readObject()
import java.util.Dictionary;
import java.util.Hashtable;
import NET.shudo.moba.lang.MobaThread;	// for loadLibrary()
import NET.shudo.moba.io.ObjectWriter;
import NET.shudo.moba.io.ObjectReader;

public final
class MetaClass {
  private static Dictionary metaClassTable = new Hashtable();
	// key  Class  instance

  private static final String STREAMABLE_CLASS_NAME =
	"NET.shudo.moba.io.Streamable";
  private static Class streamableClass = null;
  private static Object streamableClassLock = new Object();

  private static Dictionary primSigTable;

  static {
    primSigTable = new Hashtable();
    primSigTable.put("void", "V");
    primSigTable.put("boolean", "Z");
    primSigTable.put("byte", "B");
    primSigTable.put("char", "C");
    primSigTable.put("short", "S");
    primSigTable.put("int", "I");
    primSigTable.put("long", "J");
    primSigTable.put("float", "F");
    primSigTable.put("double", "D");
  }


  private transient int nFields;
  private transient int nMethods;
  private transient int nInterfaces;
  private transient MetaField[] fields = null;
  private transient MetaMethod[] methods = null;

  private transient Class thisClass = null;
  private transient Dictionary fieldNameTable = null;
  private transient Dictionary methodSelectorTable = null;

  private transient boolean hasObjRead;
  private transient boolean hasObjWrite;


  public MetaClass() {}

  /**
   *  MetaClass  Class ǽ롣
   */
  private void init(Class thisClass) {
    this.thisClass = thisClass;
    this.mySignature = this.signatureByClass(thisClass);

    Field[] flds = thisClass.getDeclaredFields();
    Constructor[] crs = thisClass.getDeclaredConstructors();
    Method[] mts = thisClass.getDeclaredMethods();
    int ncrs = crs.length;
    int nmts = mts.length;

    this.nFields = flds.length;
    this.nMethods = nmts + ncrs;
    this.nInterfaces = thisClass.getInterfaces().length;

    // MetaField 롣
    this.fields = new MetaField[nFields];
    for (int i = 0; i < nFields; i++) {
      this.fields[i] = new MetaField(thisClass, flds[i]);
    }
    // MetaMethod 롣
    this.methods = new MetaMethod[nMethods];
    for (int i = 0; i < ncrs; i++) {
      this.methods[i] = new MetaMethod(thisClass, crs[i]);
    }
    for (int i = 0; i < nmts; i++) {
      this.methods[ncrs + i] = new MetaMethod(thisClass, mts[i]);
    }

    // field, method name ɽ롣̾ -> index
    fieldNameTable = new Hashtable();
    for (int i = 0; i < this.fields.length; i++) {
      fieldNameTable.put(fields[i].name(), fields[i]);
    }
    methodSelectorTable = new Hashtable();
    for (int i = 0; i < this.methods.length; i++) {
      MetaMethod method = methods[i];
      methodSelectorTable.put(new Integer(method.ID()), method);
    }

    // writeObject(), readObject() äƤ뤫Ĵ٤롣
    this.hasObjWrite = hasOwnWriteObject0();
    this.hasObjRead = hasOwnReadObject0();

    // MetaClass ɽϿ (ѤΤ)
    synchronized(metaClassTable) {
      metaClassTable.put(this.thisClass, this);
    }
  }

  // MetaClass Ȥϥ󥹥ȥ饯ǤϤʤƤ
  public static MetaClass get(Class thisClass) {
    // native methods 饤֥Υ
    MobaThread.loadLibrary();

    MetaClass ret = null;

    synchronized(metaClassTable) {
      ret = (MetaClass)metaClassTable.get(thisClass);
      if (ret == null) {
	ret = new MetaClass();
	ret.init(thisClass);
      }
    }

    return ret;
  }
  public static MetaClass get(String className, ClassLoader loader)
	throws ClassNotFoundException {
    className = className.replace('/', '.');

    Class clazz = null;

    // native methods 饤֥Υ
    MobaThread.loadLibrary();

    try {
      if (loader != null)	// 饹ǥ
	clazz = loader.loadClass(className);
      else			// ƥ९饹ǥ
	clazz = Class.forName(className);
    }
    catch (ClassNotFoundException e) {
      e.printStackTrace();
      System.out.println();
/*
      clazz = Class.forName(className);
*/
    }

    return get(clazz);
  }
  public static MetaClass get(String className) throws ClassNotFoundException {
    return get(className, null);
  }
  public static MetaClass get(Object obj) {
    if (obj == null)  return null;
    return get(obj.getClass());
  }

  public String toString() {
    if (thisClass == null) {
      // not reached.
      return "((MetaClass)this is not initialized)";
    }

    StringBuffer buf = new StringBuffer();
    int count;

    buf.append(thisClass.toString());
    count = this.fields.length;
    if (count > 0) {
      buf.append('{');
      try {
	buf.append(field(0).name());
	for (int i = 1; i < count; i++) {
	  buf.append(',');
	  buf.append(field(i).name());
	}
      }
      catch (FieldNotFoundException e) {
	// not reached
	e.printStackTrace();  System.exit(1);
      }
      buf.append('}');
    }

    return buf.toString();
  }

  /**
   * java.lang.Class֤
   */
  public Class classClass() { return thisClass; }

  /**
   * 饹֤̾
   */
  public String getName() { return thisClass.getName(); }

  /**
   * signature ֤
   */
  public String signature() {
    return this.mySignature;
  }
  private String mySignature;

  protected static String signatureByClass(Class clazz) {
    String name = clazz.getName().replace('.', '/');

    // 
    if (name.startsWith("["))  return name;

    // ܷ
    String sig = (String)primSigTable.get(name);
    if (sig != null)  return sig;

    // (ʳ) ֥
    return "L" + name + ";";
  }


  /**
   * Υ饹Υ󥹥󥹤롣<BR>
   * κˤ newInstance(int) Ȥ
   */
  public Object newInstance() throws InstantiationException {
    Object ret = null;
    if (isArray())
      ret = null;
    else
      ret = newInstance(thisClass);
//System.out.println("MetaClass#newInstance() returns: " + ret);
    return ret;
  }
  /**
   *  () 饹Υ󥹥󥹤롣
   */
  public Object newInstance(int len, ClassLoader loader)
	throws InstantiationException {
    Object ret = null;
    if (isArray()) {
      ret = newArray(signature(), len, loader);
    }
    else
      ret = newInstance(thisClass);
    return ret;
  }
  private static native Object newInstance(Class thisClass)
	throws InstantiationException;
  private static native Object newArray(
				String sig, int len, ClassLoader loader);


  /**
   * 饹Ϳ줿ǥåĥեɤ롣
   */
  public MetaField field(int i) throws FieldNotFoundException {
    if (i < 0 || i >= this.fields.length) {
      throw new FieldNotFoundException("No such field: " + i);
    }
    return this.fields[i];
  }

  /**
   * 饹Ϳ줿ǥåĥ᥽åɤ롣
   */
  public MetaMethod method(int i) throws FieldNotFoundException {
    if (i < 0 || i >= this.methods.length) {
      throw new FieldNotFoundException("No such method: " + i);
    }
    return this.methods[i];
  }

  /**
   * Ϳ줿ե̾ĥեɤ롣
   */
  public MetaField field(String name) throws FieldNotFoundException {
    MetaField ret = (MetaField)fieldNameTable.get(name);
    if (ret == null) {
      throw new FieldNotFoundException("Field not found: " + name);
    }
    return ret;
  }

  /**
   * Ϳ줿쥯 (᥽å̾signature) ĥ᥽åɤ롣
   */
  public MetaMethod method(String name, String sig)
	throws FieldNotFoundException {
    sig = sig.replace('.', '/');
    int ID = MetaMethod.methodID(thisClass, name, sig);

    MetaMethod ret = (MetaMethod)methodSelectorTable.get(new Integer(ID));
    if (ret == null) {
      throw new FieldNotFoundException("Method not found: " + name);
    }
    return ret;
  }


  /**
   * Υ饹ʤ鿿֤<BR>
   * JDK 1.1 ʹߤ java.lang.Class#isArray() ƱΥ᥽åɡ
   */
  public boolean isArray() {
    //return thisClass.getName().startsWith("[");
    return Signature.isArray(signature());
  }

  /**
   * ֥ (or ) Υ饹ʤ鿿֤
   */
  public boolean isObjectArray() {
    return Signature.isObjectArray(signature());
  }

  /**
   * ܷ (byte, char, int, ...) ʤ鿿֤
   */
  public boolean isBaseTypeArray() {
    return Signature.isBaseTypeArray(signature());
  }


  /**
   * 饹롣<BR>
   */
  public static native Class defineClass(
	ClassLoader loader, String name, byte[] data, int offset, int length);


  /**
   * writeObject()äƤ뤫
   */
  public boolean hasOwnWriteObject() { return this.hasObjWrite; }
  private native boolean hasOwnWriteObject0();

  /**
   * readObject()äƤ뤫
   */
  public boolean hasOwnReadObject() { return this.hasObjRead; }
  private native boolean hasOwnReadObject0();


  /**
   * implementsƤ뤫Ĵ٤롣
   */
  public boolean doesImplement(Class iface) {
    // JDK 1.1 Ǥ Class#isAssignableFrom() ǼǤ롣

    if (iface == null)  return false;

    String ifaceName = iface.getName();
    Class[] ifarray = thisClass.getInterfaces();

    for (int i = 0; i < ifarray.length; i++) {
      if (ifaceName.equals(ifarray[i].getName()))  return true;
    }
    return false;
  }

  /**
   * StreamableimplementsƤ뤫Ĵ٤롣
   */
  public boolean isStreamable() {
    synchronized(streamableClassLock) {
      if (streamableClass == null) {
	try {
	  streamableClass = Class.forName(STREAMABLE_CLASS_NAME);
	}
	catch (ClassNotFoundException e) {
	  // not reached.
	  e.printStackTrace();  System.exit(1);
	}
      }
    }

    return doesImplement(streamableClass);
  }


  /**
   * super class ֤
   */
  public MetaClass superClass() {
    Class superClass = thisClass.getSuperclass();
    if (superClass == null) {
      return null;
    }
    else {
      return MetaClass.get(superClass);
    }
  }


  /**
   * constantpool count ֤(uint16)
   */
  public int constantpoolCount() {
    return ConstantPool.count(thisClass);
  }

  /**
   * access ե饰֤(uint16)
   */
  public int access() {  return thisClass.getModifiers();  }

  /**
   * եɿ֤(uint16)
   */
  public int fieldCount() {  return nFields;  }

  /**
   * ᥽åɿ֤(uint16)
   */
  public int methodCount() {  return nMethods;  }

  /**
   * interface count ֤(uint16)
   */
  public int interfaceCount() {  return nInterfaces; }


  /**
   * native methods ο֤
   */
  public int nativeMethodCount() {
    return methodCountWithAccess(Access.NATIVE);
  }

  /**
   * static methods ο֤
   */
  public int staticMethodCount() {
    return methodCountWithAccess(Access.STATIC);
  }

  /**
   * ꤷ access modifier ĥ᥽åɤο֤
   */
  private int methodCountWithAccess(int mod) {
    int count = 0;
    Method[] methods = thisClass.getDeclaredMethods();
    for (int i = 0; i < methods.length; i++) {
      int m = methods[i].getModifiers();
      if ((m & mod) == mod)  count++;
    }
    return count;
  }


  /**
   *  writeObject()
   */
  private void writeObject(ObjectWriter out) throws IOException {
//System.out.println("MetaClass#writeObject() called.");
//System.out.flush();
    out.writeUTF(this.getName());
  }

  /**
   *  readObject()
   */
  private void readObject(ObjectReader in)
	throws IOException, ClassNotFoundException {
    ClassLoader loader;
    Class clazz;
    String className = in.readUTF().replace('/', '.');

    loader = in.classLoader();
    if (loader != null)
      clazz = loader.loadClass(className);
    else
      clazz = Class.forName(className);
    this.init(clazz);
  }
}
